// feature-weighted-function.h
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: rostami@cs.nyu.com (Afshin Rostamizadeh)
//
// Implementation of the feature weighted kernel, i.e. kernel that computes
// weighted dot-products between input points.

#ifndef KERNEL_LIB_FEATURE_WEIGHTED_FUNCTION_H__
#define KERNEL_LIB_FEATURE_WEIGHTED_FUNCTION_H__

#include "kernel/lib/kernel-function.h"
#include "kernel/lib/kernel.h"
#include "kernel/lib/dense-vector.h"

namespace kernel {

template <class V> //e.g. Dense or Sparse Vector
class FeatureWeightedFunction
    : public FeaturedFunction<V, V> {
 public:
  typedef V Object;
  typedef V Feature;

  FeatureWeightedFunction(const DenseVector<typename V::ElementType> &weights,
                            int32 opt = -1)
      : weights_(weights) {
    sqrt_weights_.Resize(weights_.Size(), 0);
    for (size_t i = 0; i < sqrt_weights_.Size(); ++i) {
      CHECK(sqrt(weights_.Element(i) > 0));
      sqrt_weights_.SetElement(i, sqrt(weights_.Element(i)));
    }
  }

  ~FeatureWeightedFunction() {
  }

  Feature *GetFeature(const Object& x) const {
    CHECK(x.Size() == sqrt_weights_.Size());
    Feature *fx = new Feature;
    for (size_t i = 0; i < x.Size(); ++i)
      fx->AddElement(i, x.Element(i) * sqrt_weights_.Element(i));

    return fx;
  }

  double Product(const Feature& fx, const Feature& fy) const {
    return DotProduct(fx, fy);
  }

  double operator()(const Object& x, const Object& y) const {
    //more efficient 3-way dot product
    //CHECK(x.Size() == y.Size() && x.Size() == weights_.Size());
    double value = DotProduct(x, weights_, y);
    return value;
  }

  const uint64 Properties() const { return properties_;}

  const string Type() const { return type_;}

  bool Write(ostream &strm, const KernelWriteOptions &opts) const {
    return weights_.Write(strm);
  }

  static FeatureWeightedFunction<V> *Read(istream &strm,
                                          const KernelReadOptions &opts) {
    DenseVector<typename V::ElementType> *vec =
        DenseVector<typename V::ElementType>::Read(strm);
    FeatureWeightedFunction<V> *func = new FeatureWeightedFunction(*vec);
    delete vec;
    return func;
  }

 private:
  DenseVector<typename V::ElementType> weights_, sqrt_weights_;
  static const uint64 properties_ = kFeatured | kSymmetric | kPosDefinite;
  static const string type_;
};

template <class V>
const string FeatureWeightedFunction<V>::type_ = "feature_weighted";

} // namespace kernel

#endif  // KERNEL_LIB_FEATURE_WEIGHTED_FUNCTION_H__
