
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: rostami@google.com (Afshin Rostamizadeh)

#ifndef KERNEL_LIB_KERNEL_LEARNING_LINEARL2_H_

#include <cmath>
#include <cstdlib>
#include <vector>
#include <set>
#include <algorithm>
//#include <tr1/unordered_set>
//using std::tr1::unordered_set;

#include "kernel/lib/dense-matrix.h"
#include "kernel/lib/dense-vector.h"
#include "kernel/lib/krr.h"
#include "kernel/lib/learn-correlation.h"

//
// This file contains functions for learning l2-regularized linear
// combinations of base kernels for the kernel ridge regression (KRR)
// objective. Both general kernel combinations and specialized rank-1
// kernel combination are implemented.
//
// The learned kernel is of the form \sum_i \mu_i K_i,
// where K_i are some base kernels.
//
// The specific regularization used is ||\mu - \mu_0|| <= ker_reg,
// where \mu_0 is an additional parameter than can be used as a prior
// (e.g. uniform weights).
//
// Also has helper function for combining linear combinations of general
// kernels.
// TODO(rostami): move helper function to a new file?
//

namespace kernel {

// Learn an l2-regularized linear combination of rank-1 kernels which
// correspond to features, i.e. produce a weight for each features.
//
// INPUT:
// data - a Dense or SemiDense matrix, which each row is a training instance.
// labels - the corresponding vector of labels.
// mu0 - additional regularization parameter, can be used as a prior
// weight on each of the kernels (features in this case).
// kernel_reg - limit the norm of the generated weights.
// alg_reg - the regularization parameter that will be used by KRR problem.
// interp_param - a value between 0 and 1, values closer to 0 will
//   allow the algorithm to take bigger steps, leading to faster
//   convergence, but possibly also making the algorithm instable. A
//   good initial choice is 0.5.
// stop_tol - the stopping criteria uses this tolerance to decide when
//   to stop, a smaller value will give a more accurate solution, but
//   take longer to converge.
// max_iter - maximum number of iterations, will force function to
//   return even if stop_tol is not met.
// approx - use a rank-approx svd decomposition. (EXPERIMENTAL)
//
// OUTPUT:
// weights - the kernel (feature) weighting \mu.
// krr_vars - the dual variables from the associated KRR problem.
//
// For algorithm reference see Cortes et al. UAI 2009.
template<typename T, class M>
void MakeLinearL2Weights(const M &data,
                         const DenseVector<T> &labels,
                         const DenseVector<T> &mu0,
                         double kernel_reg,
                         double alg_reg,
                         double interp_param,
                         double stop_tol,
                         size_t max_iter,
                         int approx,
                         DenseVector<T> *weights,
                         DenseVector<T> *krr_vars) {
  CHECK(interp_param >= 0 && interp_param <= 1);
  DenseVector<T> mu(data.NumColumn(), 0);
  CHECK(mu.Size() == mu0.Size());
  DenseVector<T> alpha(data.NumRow(), 0);
  DenseVector<T> alpha_old, diff, temp;
  DenseVector<T> V(data.NumColumn(), 0);
  double V_norm, diff_norm;
  typename M::TransposeMatrix data_t;
  data.Transpose(&data_t);

  DenseVector<size_t> sample_index;
  DenseMatrix<T> K;
  M data_weighted;

  if ((approx > 0) && (approx < data.NumRow())) {
    RandomSample(0, data.NumRow(), approx, &sample_index);
  }

  if (sample_index.Size() > 0) {
    KRRApproxDualTrain(data, sample_index, labels, alg_reg, &alpha);
  } else {
    //TODO(rostami): symmetric product
    Product(data, data_t, &K);
    KRRDualTrain(K, labels, alg_reg, &alpha);
  }

  size_t iter = 0;
  do {
    alpha_old = alpha;

    //find mu for fixed alpha
    Product(data_t, alpha, &temp);
    for (size_t i = 0; i < V.Size(); ++i)
      V.SetElement(i, pow(temp.Element(i), 2));
    V_norm = sqrt(DotProduct(V, V));
    for (size_t i = 0; i < mu.Size(); ++i)
      mu.SetElement(i, mu0.Element(i) + V.Element(i) * kernel_reg / V_norm);

    //find alpha for fixed mu
    if (sample_index.Size() > 0) {
      MakeCorrelationFeatures(data, mu, &data_weighted);
      RandomSample(0, data.NumRow(), approx, &sample_index);
      KRRApproxDualTrain(data_weighted, sample_index, labels, alg_reg, &temp);
    } else {
      MakeCorrelationKernel(data, data, mu, &K);
      KRRDualTrain(K, labels, alg_reg, &temp);
    }
    for (size_t i = 0; i < alpha.Size(); ++i)
      alpha.SetElement(i,
                       interp_param * alpha.Element(i)
                       + (1 - interp_param) * temp.Element(i));

    Sum(static_cast<T>(-1), alpha, alpha_old, &diff);
    diff_norm = sqrt(DotProduct(diff, diff));
    LOG(INFO) << "iter: " << iter << " obj: "
         << linearl2_obj(alpha, labels, mu0, V, kernel_reg, alg_reg, V_norm)
         << " gap: " << diff_norm;
    ++iter;
  } while (diff_norm > stop_tol && iter < max_iter);

  *weights = mu;
  if(krr_vars != NULL)
    *krr_vars = temp;
}

// Learn an l2-regularized linear combination of general kernels. Only
// the input/output arguments that differ from the rank-1
// implementation are listed here.
//
// INPUT:
// base_kernel - an STL vector of base kernel matrices, which we wish
// to learn weights for.
template<typename T>
void MakeLinearL2Weights(const vector< DenseMatrix<T> > &base_kernels,
                         const DenseVector<T> &labels,
                         const DenseVector<T> &mu0,
                         double kernel_reg,
                         double alg_reg,
                         double interp_param,
                         double stop_tol,
                         size_t max_iter,
                         DenseVector<T> *weights,
                         DenseVector<T> *krr_vars) {
  CHECK(interp_param >= 0 && interp_param <= 1);
  CHECK(base_kernels.size() > 1);
  DenseVector<T> mu(base_kernels.size(), 0);
  CHECK(mu.Size() == mu0.Size());

  size_t m = base_kernels[0].NumRow();
  DenseVector<T> alpha(m, 0);
  DenseVector<T> alpha_old, diff, temp;
  DenseVector<T> V(base_kernels.size(), 0);
  double V_norm, diff_norm;

  DenseMatrix<T> K;
	DenseVector<T> unif(base_kernels.size(), 1);
  LinearCombination(base_kernels, unif, &K);

  KRRDualTrain(K, labels, alg_reg, &alpha);
  size_t iter = 0;
  do {
    alpha_old = alpha;

    //find mu for fixed alpha
    //V_i = alpha^T K_i alpha
    for (size_t i = 0; i < V.Size(); ++i) {
      Product(base_kernels[i], alpha, &temp);
      V.SetElement(i, DotProduct(alpha, temp));
    }
    V_norm = sqrt(DotProduct(V, V));

    for (size_t i = 0; i < mu.Size(); ++i)
      mu.SetElement(i, mu0.Element(i) + V.Element(i) * kernel_reg / V_norm);

    //find alpha for fixed mu
    LinearCombination(base_kernels, mu, &K);
    KRRDualTrain(K, labels, alg_reg, &temp);
    for (size_t i = 0; i < alpha.Size(); ++i)
      alpha.SetElement(i,
                       interp_param * alpha.Element(i)
                       + (1 - interp_param) * temp.Element(i));

    Sum(static_cast<T>(-1), alpha, alpha_old, &diff);
    diff_norm = sqrt(DotProduct(diff, diff));
    LOG(INFO) << "iter: " << iter << " obj: "
         << linearl2_obj(alpha, labels, mu0, V, kernel_reg, alg_reg, V_norm)
         << " gap: " << diff_norm;
    ++iter;
  } while (diff_norm > stop_tol && iter < max_iter);

  *weights = mu;
  if(krr_vars != NULL)
    *krr_vars = temp;
}

// Given a vector of kernel matrices k_vec, output the linear combination
// weighted by coef_vec.
template<typename T> inline
void LinearCombination(const vector< DenseMatrix<T> > &k_vec,
                       const DenseVector<T> &coef_vec,
                       DenseMatrix<T> *K) {
  CHECK(k_vec.size() > 0);
  CHECK(k_vec.size() == coef_vec.Size());
  K->Reset(k_vec[0].NumRow(), k_vec[0].NumColumn(), 0);
  for (size_t i = 0; i < k_vec.size(); ++i)
    Sum(coef_vec.Element(i), k_vec[i], static_cast<T>(1), K);
}

// This specializtion should never be used, only defined so that 
// kernel-learning-linearl1 compliles.
// TODO(rostami): a more elegant solution?
template<typename T, typename M> inline
void LinearCombination(const vector<M> &k_vec,
                       const DenseVector<T> &coef_vec,
                       M *K) {
  LOG(FATAL) << "LinearCombination non-DenseMatrix specialization used";
}

// Fill the vector v, with a uniform (without replacement) sample
// within the range [being, end).
template<typename T> inline
void RandomSample(uint64 begin, uint64 end, uint64 num_samples,
                  DenseVector<T> *v) {
  CHECK(end > begin);
  uint64 range = end - begin;
  CHECK(num_samples < range);

  vector<uint64> ids;
  set<uint64> sampled;
  for (size_t i = 0; i < num_samples; ++i) {
    int r;
    do  {
      r = rand() % range;
    } while (sampled.find(r) != sampled.end());
    sampled.insert(r);
    ids.push_back(r);
  }

  sort(ids.begin(), ids.end());
  v->Resize(ids.size(), 0);
  for (size_t i = 0; i < ids.size(); ++i)
    v->SetElement(i, ids[i]);
}

// HELPER FUNCTIONS
// Internal helper functions that are not meant to be used outside
// this file. They are called with parameters that are only readily
// available withing MakeLinearL2Weights.

// Compute the objective value of the l2-regularized multiple kernel
// learning problem.
template<typename T> inline
double linearl2_obj(const DenseVector<T> &alpha,
                    const DenseVector<T> &labels,
                    const DenseVector<T> &mu0,
                    const DenseVector<T> &V,
                    double kernel_reg,
                    double alg_reg,
                    double V_norm) {

  return -alg_reg * DotProduct(alpha, alpha)
         + 2 * DotProduct(alpha, labels)
         - DotProduct(mu0, V)
         - kernel_reg * V_norm;
}

// Compute the gradient of the objective value of the l2-regularized
// multiple kernel learning problem. Rank-1 base kernel (feature
// weighting) version.
template<class M>
double linearl2_gradnorm(const DenseVector<typename M::ElementType> &alpha,
                     M &data,
                     const DenseVector<typename M::ElementType> &labels,
                     const DenseVector<typename M::ElementType> &mu0,
                     double ker_reg,
                     double alg_reg) {

  typedef typename M::ElementType T;

  DenseMatrix<T> K0(alpha.Size(), alpha.Size(), 0);
  DenseMatrix<T> K(alpha.Size(), alpha.Size(), 0);
  typename M::TransposeMatrix data_t;
  data.Transpose(&data_t);

  DenseVector<T> V(data.NumColumn(), 0);
  DenseVector<T> temp;
  Product(data_t, alpha, &temp);
  double V_norm;
  for (size_t i = 0; i < V.Size(); ++i)
    V.SetElement(i, pow(temp.Element(i), 2));
  V_norm = sqrt(DotProduct(V, V));

  MakeCorrelationKernel(data, data, mu0, &K0);
  MakeCorrelationKernel(data, data, V, &K);
  ScalarProduct(1 / V_norm, &K);

  DenseMatrix<T> A(data.NumRow(), data.NumRow(), 0);
  Sum(1.0, K0, ker_reg, K, &A);
  for(size_t i = 0; i < A.NumRow(); ++i)
    A.SetElement(i, i, A.Element(i, i) + alg_reg);

  DenseVector<T> grad;
  Product(A, alpha, &temp);
  Sum(-1.0, temp, labels, &grad);

  return sqrt(DotProduct(grad, grad));
}

// Compute the gradient of the objective value of the l2-regularized
// multiple kernel learning problem. General base kernel version.
template<typename T>
double linearl2_gradnorm(const DenseVector<T> &alpha,
                     const vector< DenseMatrix<T> > &base_kernels,
                     const DenseVector<T> &labels,
                     const DenseVector<T> &mu0,
                     double ker_reg,
                     double alg_reg) {
  DenseMatrix<T> K0(alpha.Size(), alpha.Size(), 0);
  DenseMatrix<T> K(alpha.Size(), alpha.Size(), 0);

  DenseVector<T> V(base_kernels.size(), 0);
  DenseVector<T> temp;
  double V_norm;
  for (size_t i = 0; i < V.Size(); ++i) {
    Product(base_kernels[i], alpha, &temp);
    V.SetElement(i, DotProduct(alpha, temp));
  }
  V_norm = sqrt(DotProduct(V, V));

  LinearCombination(base_kernels, mu0, &K0);
  LinearCombination(base_kernels, V, &K);
  ScalarProduct(1 / V_norm, &K);

  DenseMatrix<T> A(alpha.Size(), alpha.Size(), 0);
  Sum(1.0, K0, ker_reg, K, &A);
  for(size_t i = 0; i < A.NumRow(); ++i)
    A.SetElement(i, i, A.Element(i, i) + alg_reg);

  DenseVector<T> grad;
  Product(A, alpha, &temp);
  Sum(-1.0, temp, labels, &grad);

  return sqrt(DotProduct(grad, grad));
}

} //namespace kernel

#define KERNEL_LIB_KERNEL_LEARNING_LINEARL2_H_

#endif  // KERNEL_LIB_KERNEL_LEARNING_LINEARL2_H_
