// rational-kernel.h
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: allauzen@cs.nyu.edu (Cyril Allauzen)
//
// \file
// Rational kernel representation.

#ifndef KERNEL_LIB_RATIONAL_KERNEL_H__
#define KERNEL_LIB_RATIONAL_KERNEL_H__

#include "kernel/lib/kernel.h"
#include <fst/fst.h>
#include "kernel/lib/basic-kernel.h"
#include "kernel/lib/rational-function.h"

namespace kernel {

using namespace fst;

typedef Fst<LogArc> LogFst;

template <class A>
class RationalKernel :
      public BasicKernel< RationalFunction<A>,
                          BasicCache< Fst<A> , VectorFst<A> > > {
 public:
  typedef RationalFunction<A> Function;
  typedef typename Function::Object Object;
 private:
  typedef BasicCache< Fst<A>, VectorFst<A> > Cache;

 public:
  RationalKernel(const Fst<A> &fst, int32 opt = -1, bool normalize = true,
                 uint32 cache_type = kCacheFeature | kCacheValue)
      : BasicKernel<Function, Cache>(new Function(fst, opt),
                                     new Cache(cache_type),
                                     normalize) {
    BasicKernel<Function, Cache>::Impl()->SetType("rational");
  }

  RationalKernel(const BasicKernel<Function, Cache> &kernel)
      : BasicKernel<Function, Cache>(kernel) {
    BasicKernel<Function, Cache>::Impl()->SetType("rational");
  }

  static RationalKernel<A> *Read(istream &strm, const KernelReadOptions &opts) {
    KernelReadOptions ropts(opts);
    ropts.type = "rational";
    BasicKernelImpl<Function, Cache>* impl =
      BasicKernelImpl<Function, Cache>::Read(strm, ropts);
    return impl ? new RationalKernel<A>(impl) : 0;
  }

  static RationalKernel<A> *Read(const string &filename) {
    if (!filename.empty()) {
      ifstream strm(filename.c_str());
      if (!strm) {
        LOG(ERROR) << "RationalKernel::Read: Can't open file: " << filename;
        return 0;
      }
      return Read(strm, KernelReadOptions(filename));
    } else {
      return Read(std::cin, KernelReadOptions("standard input"));
    }
  }

 private:
  explicit RationalKernel(BasicKernelImpl<Function, Cache> *impl) :
      BasicKernel<Function, Cache>(impl) {
    BasicKernel<Function, Cache>::Impl()->SetType("rational");
  }
};

template <class A>
class UnfactoredRationalKernel :
      public BasicKernel< UnfactoredRationalFunction<A>,
                          BasicCache< Fst<A> , VectorFst<A> > > {
 public:
  typedef UnfactoredRationalFunction<A> Function;
  typedef typename Function::Object Object;
 private:
  typedef BasicCache< Fst<A>, VectorFst<A> > Cache;

 public:
  UnfactoredRationalKernel(const Fst<A> &fst,
                           int32 opt = -1,
                           bool normalize = true,
                           bool symmetric = true,
                           uint32 cache_type = kCacheValue)
      : BasicKernel<Function, Cache>(new Function(fst, opt),
                                     new Cache(cache_type),
                                     normalize) {
    BasicKernel<Function, Cache>::Impl()->SetType("unfactored_rational");
    BasicKernel<Function, Cache>::Impl()->SetProperties(
        symmetric ? kSymmetric : kAsymmetric, kSymmetric | kAsymmetric);
  }

  UnfactoredRationalKernel(const BasicKernel<Function, Cache> &kernel)
      : BasicKernel<Function, Cache>(kernel) {
    BasicKernel<Function, Cache>::Impl()->SetType("unfactored_rational");
  }

  static UnfactoredRationalKernel<A> *Read(istream &strm,
                                           const KernelReadOptions &opts) {
    KernelReadOptions ropts(opts);
    ropts.type = "unfactored_rational";
    BasicKernelImpl<Function, Cache>* impl =
      BasicKernelImpl<Function, Cache>::Read(strm, ropts);
    return impl ? new UnfactoredRationalKernel<A>(impl) : 0;
  }

  static UnfactoredRationalKernel<A> *Read(const string &filename) {
    if (!filename.empty()) {
      ifstream strm(filename.c_str());
      if (!strm) {
        LOG(ERROR) << "UnfactoredRationalKernel::Read: Can't open file: "
                   << filename;
        return 0;
      }
      return Read(strm, KernelReadOptions(filename));
    } else {
      return Read(std::cin, KernelReadOptions("standard input"));
    }
  }

 private:
  explicit UnfactoredRationalKernel(BasicKernelImpl<Function, Cache> *impl) :
      BasicKernel<Function, Cache>(impl) {
    BasicKernel<Function, Cache>::Impl()->SetType("unfactored_rational");
  }
};


}  // namespace kernel

#endif  // KERNEL_LIB_RATIONAL_KERNEL_H__
